#!/usr/bin/perl
#
#  Author: Preslav Nakov
#  
#  Description: Scores subtask B of SemEval-2015 Task 3.
#
#  Version: 1.0
#
#  Last modified: October 2, 2014
#
#
#  Use:
#     SemEval2015-task3-scorer-subtaskB.pl <GOLD_FILE> <PREDICTIONS_FILE>
#
#  Example use:
#     SemEval2015-task3-scorer-subtaskB.pl CQA-QL-devel-gold-yn.txt CQA-QL-devel-predicted-yn-good.txt > SemEval2015-task3-scores-subtaskB.txt
#     SemEval2015-task3-scorer-subtaskB.pl CQA-QL-devel-gold-yn.txt CQA-QL-devel-predicted-yn-majority-baseline.txt > SemEval2015-task3-scores-subtaskB-majority-baseline.txt
#
#  Description:
#     The scorer takes as input a proposed classification file and an answer key file.
#     Both files should contain one prediction per line in the format "<Question_ID>	<Label>"
#     with a TAB as a separator, e.g.,
#           Q2601	Yes
#			Q2602	No
#			Q2603	Unsure
#               ...
#     The files have to be sorted in the same way, i.e., their <Question_ID> should match for each line!
#     Repetitions of IDs are not allowed in either of the files.
#
#     The scorer calculates and outputs the following statistics:
#        (1) confusion matrix, which shows
#			- the count for each gold/predicted pair
#           - the sums for each row/column: -SUM-
#        (2) accuracy
#        (3) precision (P), recall (R), and F1-score for each label
#        (4) micro-averaged P, R, F1
#        (5) macro-averaged P, R, F1
#
#     
#     The official score is the macro-averaged F1-score.
#

use warnings;
use strict;
use utf8;


###################
###   GLOBALS   ###
###################

my %confMatrix   = ();
my @allLabels    = ('Yes', 'No', 'Uns.');
my %labelMapping = ('Yes'=>'Yes', 'No'=>'No', 'Unsure'=>'Uns.');


################
###   MAIN   ###
################

### 1. Check parameters
die "Usage: $0 <GOLD_FILE> <PREDICTIONS_FILE>\n" if ($#ARGV != 1);
my $GOLD_FILE        = $ARGV[0];
my $PREDICTIONS_FILE = $ARGV[1];

### 2. Open the files
open GOLD, $GOLD_FILE or die "Error opening $GOLD_FILE!";
open PREDICTED, $PREDICTIONS_FILE or die "Error opening $PREDICTIONS_FILE!";
binmode(GOLD, ":utf8");
binmode(PREDICTED, ":utf8");

### 3. Collect the statistics
my %QIDs = ();
while (<GOLD>) {
	
	# 3.1. Get the GOLD QID and label
	# Q2601	Yes
	die "Wrong file format!" if (!/^Q([0-9]+)\t(Yes|No|Unsure)[\n\r]*$/);
	my ($goldQID, $goldLabel) = ($1, $2);

	# 3.2. Get the PREDICTED QID and label
	# Q2601	Unsure
	die "The file $PREDICTIONS_FILE is shorter!" if (!($_ = <PREDICTED>));
	die "Wrong file format!" if (!/^Q([0-9]+)\t(Yes|No|Unsure)[\n\r]*$/);
	my ($predictedQID, $predictedLabel) = ($1, $2);

	# 3.3. Make sure IDs match
	die "Question IDs differ: gold='$goldQID' predicted='$predictedQID'" if ($predictedQID ne $goldQID);

	# 3.4. Make sure this Question ID was not seen already
	die "Repetition of Question ID: $goldQID\n" if (defined $QIDs{$goldQID});
	$QIDs{$goldQID}++;

	# 3.5. Update the statistics
	$confMatrix{$labelMapping{$predictedLabel}}{$labelMapping{$goldLabel}}++;
}

### 4. Evaluation
print "\n<<< I. EVALUATION >>>\n\n";
my $officialScore = &evaluate(\@allLabels, \%confMatrix);

### 5. Output the official score
print "\n<<< II. OFFICIAL SCORE >>>\n";
printf "\nMACRO-averaged F1: %6.2f%s", $officialScore, "%\n";

### 6. Close the files
close GOLD or die;
close PREDICTED or die;


################
###   SUBS   ###
################

sub evaluate() {
	my ($allLabels, $confMatrix) = @_;

	### 0. Calculate the horizontal and vertical sums
	my %allLabelsProposed = ();
	my %allLabelsAnswer   = ();
	my ($cntCorrect, $cntTotal) = (0, 0);
	foreach my $labelGold (@{$allLabels}) {
		foreach my $labelProposed (@{$allLabels}) {
			$$confMatrix{$labelProposed}{$labelGold} = 0
				if (!defined($$confMatrix{$labelProposed}{$labelGold}));
			$allLabelsProposed{$labelProposed} += $$confMatrix{$labelProposed}{$labelGold};
			$allLabelsAnswer{$labelGold} += $$confMatrix{$labelProposed}{$labelGold};
			$cntTotal += $$confMatrix{$labelProposed}{$labelGold};
		}
		$cntCorrect += $$confMatrix{$labelGold}{$labelGold};
	}

	### 1. Print the confusion matrix heading
	print "Confusion matrix:\n";
	print "       ";
	foreach my $label (@{$allLabels}) {
		printf " %4s", $label;
	}
	print " <-- classified as\n";
	print "      +";
	foreach (@{$allLabels}) {
		print "-----";
	}
	print "+ -SUM-\n";

	### 2. Print the rest of the confusion matrix
	my $freqCorrect = 0;
	foreach my $labelGold (@{$allLabels}) {

		### 2.1. Output the short relation label
		printf " %4s |", $labelGold;

		### 2.2. Output a row of the confusion matrix
		foreach my $labelProposed (@{$allLabels}) {
			printf "%4d ", $$confMatrix{$labelProposed}{$labelGold};
		}

		### 2.3. Output the horizontal sums
		printf "| %4d\n", $allLabelsAnswer{$labelGold};
	}
	print "      +";
	foreach (@{$allLabels}) {
		print "-----";
	}
	print "+\n";
	
	### 3. Print the vertical sums
	print " -SUM- ";
	foreach my $labelProposed (@{$allLabels}) {
		printf "%4d ", $allLabelsProposed{$labelProposed};
	}
	print "\n\n";

	### 5. Output the accuracy
	my $accuracy = 100.0 * $cntCorrect / $cntTotal;
	printf "%s%d%s%d%s%5.2f%s", 'Accuracy (calculated for the above confusion matrix) = ', $cntCorrect, '/', $cntTotal, ' = ', $accuracy, "\%\n";

	### 8. Output P, R, F1 for each relation
	my ($macroP, $macroR, $macroF1) = (0, 0, 0);
	my ($microCorrect, $microProposed, $microAnswer) = (0, 0, 0);
	print "\nResults for the individual labels:\n";
	foreach my $labelGold (@{$allLabels}) {

		### 8.3. Calculate P/R/F1
		my $P  = (0 == $allLabelsProposed{$labelGold}) ? 0
				: 100.0 * $$confMatrix{$labelGold}{$labelGold} / $allLabelsProposed{$labelGold};
		my $R  = (0 == $allLabelsAnswer{$labelGold}) ? 0
				: 100.0 * $$confMatrix{$labelGold}{$labelGold} / $allLabelsAnswer{$labelGold};
		my $F1 = (0 == $P + $R) ? 0 : 2 * $P * $R / ($P + $R);

		printf "%10s%s%4d%s%4d%s%6.2f", $labelGold,
			" :    P = ", $$confMatrix{$labelGold}{$labelGold}, '/', $allLabelsProposed{$labelGold}, ' = ', $P;

		printf"%s%4d%s%4d%s%6.2f%s%6.2f%s\n", 
		  	 "%     R = ", $$confMatrix{$labelGold}{$labelGold}, '/', $allLabelsAnswer{$labelGold},   ' = ', $R,
			 "%     F1 = ", $F1, '%';

		### 8.5. Accumulate statistics for micro/macro-averaging
		if ($labelGold ne '_Other') {
			$macroP  += $P;
			$macroR  += $R;
			$macroF1 += $F1;
			$microCorrect += $$confMatrix{$labelGold}{$labelGold};
			$microProposed += $allLabelsProposed{$labelGold};
			$microAnswer += $allLabelsAnswer{$labelGold};
		}
	}

	### 9. Output the micro-averaged P, R, F1
	my $microP  = (0 == $microProposed)    ? 0 : 100.0 * $microCorrect / $microProposed;
	my $microR  = (0 == $microAnswer)      ? 0 : 100.0 * $microCorrect / $microAnswer;
	my $microF1 = (0 == $microP + $microR) ? 0 :   2.0 * $microP * $microR / ($microP + $microR);
	print "\nMicro-averaged result:\n";
	printf "%s%4d%s%4d%s%6.2f%s%4d%s%4d%s%6.2f%s%6.2f%s\n",
		      "P = ", $microCorrect, '/', $microProposed, ' = ', $microP,
		"%     R = ", $microCorrect, '/', $microAnswer, ' = ', $microR,
		"%     F1 = ", $microF1, '%';

	### 10. Output the macro-averaged P, R, F1
	my $distinctLabelsCnt = $#{$allLabels}+1; 

	$macroP  /= $distinctLabelsCnt; # first divide by the number of non-Other categories
	$macroR  /= $distinctLabelsCnt;
	$macroF1 /= $distinctLabelsCnt;
	print "\nMACRO-averaged result:\n";
	printf "%s%6.2f%s%6.2f%s%6.2f%s\n\n\n\n", "P = ", $macroP, "%\tR = ", $macroR, "%\tF1 = ", $macroF1, '%';

	### 11. Return the official score
	return $macroF1;
}
